# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest

from enum import Enum

from hamcrest import *

from samitorch.factories.factories import ModelFactory

from samitorch.factories.parsers import ModelConfigurationParserFactory

from samitorch.factories.enums import UNetModels, ResNetModels


class IncorrectModels(Enum):
    Incorrect = "Incorrect"


class ModelFactoryTest(unittest.TestCase):
    RESNET_MODEL = ResNetModels.ResNet18
    RESNET_CONFIGURATION_PATH = "samitorch/configs/resnet3d.yaml"
    UNET_MODEL = UNetModels.UNet3D
    UNET_CONFIGURATION_PATH = "samitorch/configs/unet3d.yaml"
    UNKNOWN_MODEL = IncorrectModels.Incorrect
    INVALID_CONFIG_PATH = "samitorch/configs/invalid.yaml"

    def setUp(self):
        self.model_factory = ModelFactory()
        self.configurationParserFactory = ModelConfigurationParserFactory()
        self.resnet_config = self.configurationParserFactory.parse(self.RESNET_CONFIGURATION_PATH)
        self.unet_config = self.configurationParserFactory.parse(self.UNET_CONFIGURATION_PATH)

    def test_should_instantiate_restnet_model(self):
        model = self.model_factory.create_model(self.RESNET_MODEL, self.resnet_config)
        assert_that(model, is_not(None))

    def test_should_instantiate_unet_model(self):
        model = self.model_factory.create_model(self.UNET_MODEL, self.unet_config)
        assert_that(model, is_not(None))

    def test_should_fail_with_unknown_model(self):
        model_factory = ModelFactory()
        assert_that(calling(model_factory.create_model).with_args(self.UNKNOWN_MODEL, self.resnet_config),
                    raises(ValueError))



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch


class TestModelHelper(object):

    def __init__(self, model, criterion, optimizer):
        self._model = model
        self._criterion = criterion
        self._optimizer = optimizer

    def _train_step(self, inputs: tuple):
        """
        Run a training step on model for a given batch of data.

        Parameters of the model accumulate gradients and the optimizer performs
        a gradient update on the parameters

        Args:
            inputs (tuple): A 2-element tuple of inputs and labels, to be fed to the model.
        """
        self._model.train()
        self._optimizer.zero_grad()
        inputs, targets = inputs
        preds = self._model.forward(inputs)
        loss = self._criterion(preds, targets)
        loss.backward()
        self._optimizer.step()

    def _forward_step(self, inputs: list):
        """
        Run a forward step on the model.

        Args:
            inputs (list): Model's inputs.

        Returns:
            torch.Tensor: Model's predictions (output).

        """
        self._model.eval()
        with torch.no_grad():
            return self._model(inputs)

    def _var_change_helper(self, vars_change: bool, inputs: tuple, params: list = None):
        """
        Check if given variables (params) change or not during training
        If parameters (params) aren't provided, check all parameters.

        Args:
            vars_change (bool): A flag which controls the check for change or not change.
            inputs (tuple): A 2-element tuple of inputs and labels, to be fed to the model.
            params (list): List of parameters of form (name, variable), optional.

        Raises:
            ValueError: if the model's variables do not behave accordingly to vars_change parameter.
        """
        if params is None:
            # get a list of params that are allowed to change
            params = [np for np in self._model.named_parameters() if np[1].requires_grad]

        # take a copy
        initial_params = [(name, p.clone()) for (name, p) in params]

        # run a training step
        self._train_step(inputs)

        # check if variables have changed
        for (_, p0), (name, p1) in zip(initial_params, params):
            try:
                if vars_change:
                    assert not torch.equal(p0, p1)
                else:
                    assert torch.equal(p0, p1)
            except AssertionError:
                raise ValueError(  # error message
                    "{var_name} {msg}".format(
                        var_name=name,
                        msg='did not change' if vars_change else 'changed'
                    )
                )

    def assert_vars_change(self, inputs: tuple, params: list = None):
        """
        Make sure that the given parameters (params) DO change during training
        If parameters (params) aren't provided, check all parameters.

        Args:
              inputs (tuple): A 2-element tuple of inputs and labels, to be fed to the model.
              params (list): List of parameters of form (name, variable), optional.

        Raises:
              ValueError: if the model's variables do not change.
        """

        self._var_change_helper(True, inputs, params)

    def assert_var_same(self, inputs: tuple, params: list = None):
        """
        Make sure that the given parameters (params) DO NOT change during training
        If parameters (params) aren't provided, check all parameters.

        Args:
              inputs (tuple): A 2-element tuple of inputs and labels, to be fed to the model.
              params (list): List of parameters of form (name, variable), optional.

        Raises:
              ValueError: if the model's variables change.
        """
        self._var_change_helper(False, inputs, params)

    @staticmethod
    def assert_not_nan(tensor):
        """
        Make sure there are no NaN values in the given tensor.

        Args:
            tensor (:obj:`torch.Tensor`): Input tensor.

        Raises:
            ValueError: If one or more NaN values occur in the given tensor.
        """
        try:
            assert not torch.isnan(tensor).byte().any()
        except AssertionError:
            raise ValueError("There was a NaN value in tensor.")

    @staticmethod
    def assert_never_inf(tensor):
        """
        Make sure there are no Inf values in the given tensor.

        Args:
            tensor (:obj:`torch.Tensor`): Input tensor.

        Raises:
            ValueError: If one or more Inf values occur in the given tensor.
        """
        try:
            assert torch.isfinite(tensor).byte().any()
        except AssertionError:
            raise ValueError("There was an Inf value in tensor")



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
import torch

from samitorch.factories.enums import ActivationLayers, PoolingLayers


class Configuration(metaclass=abc.ABCMeta):
    """
    A standard Configuration object.
    """

    @abc.abstractmethod
    def __init__(self) -> None:
        """
        Initialize a Configuration object.

        An implementation of Configuration object must take a dictionary as parameter. The __init__ method places the
        content of the dictionary into Python variables.

        Examples::
            self._attribute = config["attribute"]
            self._attribute = config.get("attribute", default=None)
        """
        pass


class DatasetConfiguration(Configuration, metaclass=abc.ABCMeta):
    """
    A base class for storing dataset configuration as an object.
    """

    @abc.abstractmethod
    def __init__(self, config: dict) -> None:
        super(DatasetConfiguration, self).__init__()
        pass


class ModelConfiguration(Configuration, metaclass=abc.ABCMeta):
    """
    A base class for storing model configuration as an object.
    """

    @abc.abstractmethod
    def __init__(self, config: dict) -> None:
        super(ModelConfiguration, self).__init__()
        pass


class TrainingConfiguration(Configuration, metaclass=abc.ABCMeta):
    """
    A base class for storing training configuration as an object.
    """

    @abc.abstractmethod
    def __init__(self, config: dict) -> None:
        super(TrainingConfiguration, self).__init__()
        pass


class MetricConfiguration(Configuration, metaclass=abc.ABCMeta):
    """
    A base class for storing metric configuration as an object.
    """

    @abc.abstractmethod
    def __init__(self, config: dict) -> None:
        super(MetricConfiguration, self).__init__()
        pass


class VariableConfiguration(Configuration, metaclass=abc.ABCMeta):
    """
    A base class for storing various variables using during training as an object.
    """

    @abc.abstractmethod
    def __init__(self, config: dict) -> None:
        super(VariableConfiguration, self).__init__()
        pass


class LoggerConfiguration(Configuration, metaclass=abc.ABCMeta):
    """
    A base class for storing logger attributes.
    """

    @abc.abstractmethod
    def __init__(self, config: dict) -> None:
        super(LoggerConfiguration, self).__init__()
        pass


class RunningConfiguration(Configuration):
    def __init__(self, config: dict):
        super(RunningConfiguration, self).__init__()
        self._opt_level = config.get("opt_level", default="00")
        self._num_workers = config.get("num_workers", default=0)
        self._local_rank = config.get("local_rank", default=0)
        self._sync_batch_norm = config.get("sync_batch_norm", None)
        self._keep_batch_norm_fp32 = config.get("keep_batch_norm_fp32", None)
        self._loss_scale = config.get("loss_scale", None)
        self._num_gpus = config.get("num_gpus", 0)
        self._log_path = config.get("log", "/tmp")
        self._device = torch.device("cuda:" + str(self._local_rank)) if torch.cuda.is_available() else torch.device(
            "cpu")

    @property
    def opt_level(self) -> str:
        """
        The optimization level.
            - 00: FP32 training
            - 01: Mixed Precision (recommended)
            - 02: Almost FP16 Mixed Precision
            - 03: FP16 Training.

        Returns:
            str: The optimization level.
        """
        return self._opt_level

    @property
    def num_workers(self) -> int:
        """
        The number of data loading workers (default: 4).

        Returns:
            int: The number of parallel threads.
        """
        return self._num_workers

    @property
    def local_rank(self) -> int:
        """
        The local rank of the distributed node.

        Returns:
            int: The local rank.
        """
        return self._local_rank

    @property
    def sync_batch_norm(self) -> bool:
        """
        Enables the APEX sync of batch normalization.

        Returns:
            bool: Whether if synchronization is enabled or not.
        """
        return self._sync_batch_norm

    @property
    def keep_batch_norm_fp32(self) -> bool:
        """
        Whether to keep the batch normalization in 32-bit floating point (Mixed Precision).

        Returns:
            bool: True will keep 32-bit FP batch norm, False will convert it to 16-bit FP.
        """
        return self._keep_batch_norm_fp32

    @property
    def loss_scale(self) -> str:
        """
        The loss scale in Mixed Precision training.

        Returns:
            The loss scale.
        """
        return self._loss_scale

    @property
    def num_gpus(self) -> int:
        """
        The number of GPUs on the Node to be used for training.

        Returns:
            int: The number of allowed GPU to be used for training.
        """
        return self._num_gpus

    @property
    def log_path(self) -> str:
        """
        The path where logs are saved.

        Returns:
            str: The path to log directory.
        """
        return self._log_path

    @property
    def device(self) -> torch.device:
        """
        Get the device where Tensors are going to be transfered.

        Returns:
            :obj:`torch.device`: A Torch Device object.
        """
        return self._device


class DiceMetricConfiguration(MetricConfiguration):
    def __init__(self, config):
        super(DiceMetricConfiguration, self).__init__(config)

        self._num_classes = config["num_classes"]
        self._reduction = config["reduction"]
        self._ignore_index = config["ignore_index"]
        self._average = config["average"]

    @property
    def num_classes(self) -> int:
        """
        int: The number of classes of the problem.
        """
        return self._num_classes

    @property
    def reduction(self) -> str:
        """
        str: The reduction method (e.g. "mean").
        """
        return self._reduction

    @property
    def ignore_index(self) -> int:
        """
        int: The ignored index in a problem.
        """
        return self._ignore_index

    @property
    def average(self) -> str:
        """
        str: The average method in case the metric uses a Confusion Matrix
        """
        return self._average


class UNetModelConfiguration(ModelConfiguration):
    """
    Configuration properties for a UNet model.
    """

    def __init__(self, config: dict) -> None:
        """
        Instantiate a UnetModelConfiguration object regrouping every UNet3D model's hyper-parameters.

        Args:
            config (dict): A dictionary containing model's hyper-parameters.
        """
        super(UNetModelConfiguration, self).__init__(config)

        self._feature_maps = config["feature_maps"]
        self._in_channels = config["in_channels"]
        self._out_channels = config["out_channels"]
        self._num_levels = config["num_levels"]
        self._conv_kernel_size = config["conv_kernel_size"]
        self._pool_kernel_size = config["pool_kernel_size"]
        self._pooling_type = PoolingLayers(config["pooling_type"])
        self._num_groups = config["num_groups"]
        self._padding = config["padding"]
        self._activation = ActivationLayers(config["activation"])
        self._interpolation = config["interpolation"]
        self._scale_factor = config["scale_factor"]

    @property
    def feature_maps(self) -> int:
        """
        int: Number of feature maps of first UNet level.
        """
        return self._feature_maps

    @property
    def in_channels(self) -> int:
        """
        int: Number of input channels (modality).
        """
        return self._in_channels

    @property
    def out_channels(self) -> int:
        """
        int: Number of output channels.
        """
        return self._out_channels

    @property
    def num_levels(self) -> int:
        """
        int: Number of levels in the UNet architecture.
        """
        return self._num_levels

    @property
    def conv_kernel_size(self) -> int:
        """
        int: The convolution kernel size as integer.
        """
        return self._conv_kernel_size

    @property
    def pool_kernel_size(self) -> int:
        """
        int: The pooling kernel size as integer.
        """
        return self._pool_kernel_size

    @property
    def pooling_type(self) -> PoolingLayers:
        """
        :obj:`samitorch.models.enums.PoolingLayers`: The pooling type.
        """
        return self._pooling_type

    @property
    def num_groups(self) -> int:
        """
        int: The number of groups in group normalization.
        """
        return self._num_groups

    @property
    def padding(self) -> tuple:
        """
        tuple: The padding size of each dimension.
        """
        return self._padding

    @property
    def activation(self) -> ActivationLayers:
        """
        :obj:`samitorch.models.enums.ActivationLayers`: The activation function as a string.
        """
        return self._activation

    @property
    def interpolation(self) -> bool:
        """
        bool: Whether the decoder is doing interpolation (True) or transposed convolution (False).
        """
        return self._interpolation

    @property
    def scale_factor(self) -> tuple:
        """
        tuple: The scale factor (or stride in the transposed convolution) in the decoding path.
        """
        return self._scale_factor


class ResNetModelConfiguration(ModelConfiguration):
    """
    Configuration properties for a ResNet model.
    """

    def __init__(self, config: dict) -> None:
        """
        Instantiate a ResNetModelConfiguration object regrouping every ResNet3D model's hyper-parameters.

        Args:
            config (dict): A dictionary containing model's hyper-parameters.
        """
        super(ResNetModelConfiguration, self).__init__(config)

        self._in_channels = config["in_channels"]
        self._out_channels = config["out_channels"]
        self._num_groups = config["num_groups"]
        self._conv_groups = config["conv_groups"]
        self._width_per_group = config["width_per_group"]
        self._padding = config["padding"]
        self._activation = ActivationLayers(config["activation"])
        self._zero_init_residual = config["zero_init_residual"]
        self._replace_stride_with_dilation = config["replace_stride_with_dilation"]

    @property
    def in_channels(self) -> int:
        """
        int: Number of input channels (modality).
        """
        return self._in_channels

    @property
    def out_channels(self) -> int:
        """
        int: Number of output channels.
        """
        return self._out_channels

    @property
    def num_groups(self) -> int:
        """
        int: The number of groups in group normalization.
        """
        return self._num_groups

    @property
    def conv_groups(self) -> int:
        """
        int: The number of groups that control the connections between inputs and outputs convolutions.
        """
        return self._conv_groups

    @property
    def width_per_group(self) -> int:
        """
        int: The width of convolution groups.

        Notes:
            Variable used in the width calculation formula of ResNet convolutional groups: `width = int(planes * (width_per_group / 64.)) * groups`
            And then used in torch.nn.conv3d operations as in_channels and out_channels parameters.
        """
        return self._width_per_group

    @property
    def padding(self) -> tuple:
        """
        tuple: The padding size of each dimensions.
        """
        return self._padding

    @property
    def activation(self) -> ActivationLayers:
        """
        :obj:`samitorch.models.enums.ActivationLayers`: The activation function as a string.
        """
        return self._activation

    @property
    def zero_init_residual(self) -> str:
        """
        bool:  Zero-initialize the last batch normalization layer in each residual branch.
        """
        return self._zero_init_residual

    @property
    def replace_stride_with_dilation(self):
        """
        tuple: A tuple of boolean where each element in the tuple indicates if we should replace
            the 2x2 stride with a dilated convolution instead
        """
        return self._replace_stride_with_dilation



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import random

from typing import Callable, List, Optional, Tuple

from torch.utils.data.dataset import Dataset

from samitorch.utils.utils import extract_file_paths
from samitorch.inputs.sample import Sample
from samitorch.inputs.transformers import ToNumpyArray, PadToPatchShape
from torchvision.transforms.transforms import Compose

from samitorch.utils.slice_builder import SliceBuilder


class NiftiPatchDataset(Dataset):
    """
    Create a dataset of patches in PyTorch for reading NiFTI files and slicing them into fixed shape.
    """

    def __init__(self, source_dir: str, target_dir: str, patch_shape: Tuple[int, int, int, int],
                 step: Tuple[int, int, int, int], dataset_id: int = None,
                 transform: Optional[Callable] = None) -> None:
        """
        Dataset initializer.

        Args:
            source_dir (str): Path to source images.
            target_dir (str): Path to target (labels) images.
            transform (Callable): transform to apply to both source and target images.
        """
        self._source_dir, self._target_dir = source_dir, target_dir
        self._source_paths, self._target_paths = extract_file_paths(source_dir), extract_file_paths(target_dir)
        self._dataset_id = dataset_id
        self._transform = transform

        if len(self._source_paths) != len(self._target_paths) or len(self._source_paths) == 0:
            raise ValueError("Number of source and target images must be equal and non-zero.")

        pre_transforms = Compose([ToNumpyArray(), PadToPatchShape(patch_shape, step)])

        self._images = list()
        self._labels = list()

        for idx in range(len(self._source_paths)):
            source_path, target_path = self._source_paths[idx], self._target_paths[idx]
            sample = Sample(x=source_path, y=target_path, dataset_id=self._dataset_id, is_labeled=True)
            transformed_sample = pre_transforms(sample)
            self._images.append(transformed_sample.x)
            self._labels.append(transformed_sample.y)

        self._slice_builder = SliceBuilder(self._images[0].shape, patch_size=patch_shape, step=step)

        slices = []
        for i in range(len(self._images)):
            slices.append(list(
                filter(lambda slice: np.count_nonzero(self._images[i][slice]) > 0, self._slice_builder.image_slices)))

        self._slices = [j for sub in slices for j in sub]
        self._slices = np.array(self._slices)
        self._num_patches = self._slices.shape[0]

    @property
    def slices(self):
        return self._slices

    @slices.setter
    def slices(self, slices):
        self._slices = slices

    @property
    def num_patches(self):
        return self._num_patches

    @num_patches.setter
    def num_patches(self, num_patches):
        self._num_patches = num_patches

    def __len__(self):
        return self._num_patches

    def __getitem__(self, idx: int):
        id = random.randint(0, len(self._images) - 1)
        slice = self._slices[idx]
        img = self._images[id]
        label = self._labels[id]
        x, y = img[tuple(slice)], label[tuple(slice)]

        sample = Sample(x=x, y=y, dataset_id=self._dataset_id, is_labeled=True)

        if self._transform is not None:
            sample = self._transform(sample)
        return sample


class NiftiDataset(Dataset):
    """
    Create a dataset class in PyTorch for reading NIfTI files.
    """

    def __init__(self, source_dir: str, target_dir: str, dataset_id: int = None,
                 transform: Optional[Callable] = None) -> None:
        """
        Dataset initializer.

        Args:
            source_dir (str): Path to source images.
            target_dir (str): Path to target (labels) images.
            transform (Callable): transform to apply to both source and target images.
        """
        self._source_dir, self._target_dir = source_dir, target_dir
        self._source_paths, self._target_paths = extract_file_paths(source_dir), extract_file_paths(target_dir)
        self._dataset_id = dataset_id
        self._transform = transform

        if len(self._source_paths) != len(self._target_paths) or len(self._source_paths) == 0:
            raise ValueError("Number of source and target images must be equal and non-zero.")

    def __len__(self):
        return len(self._source_paths)

    def __getitem__(self, idx: int):
        source_path, target_path = self._source_paths[idx], self._target_paths[idx]
        sample = Sample(x=source_path, y=target_path, dataset_id=self._dataset_id, is_labeled=True)

        if self._transform is not None:
            sample = self._transform(sample)
        return sample


class MultimodalDataset(Dataset):
    """
    Base class for Multimodal NifTI Dataset.
    """

    def __init__(self, source_dirs: List[str], target_dirs: List[str], dataset_id: int = None,
                 transform: Optional[Callable] = None) -> None:
        """
        Multimodal Dataset initializer.

        Args:
            source_dirs (List[str]): Paths to source images.
            target_dirs (List[str]): paths to target (labels) images.
            transform (Callable): transform to apply to both source and target images.w
        """
        self._source_dirs, self._target_dirs = source_dirs, target_dirs
        self._source_paths, self._target_paths = [extract_file_paths(sd) for sd in source_dirs], [extract_file_paths(td) for td
                                                                                                  in
                                                                                                  target_dirs]
        self._dataset_id = dataset_id
        self._transform = transform

        if any([len(self._source_paths[0]) != len(sfn) for sfn in self._source_paths]) or \
                any([len(self._target_paths[0]) != len(tfn) for tfn in self._target_paths]) or \
                len(self._source_paths[0]) != len(self._target_paths[0]) or \
                len(self._source_paths[0]) == 0:
            raise ValueError(f'Number of source and target images must be equal and non-zero')

    def __len__(self):
        return len(self._source_paths[0])

    def __getitem__(self, idx: int):
        source_image, target_image = [source_path[idx] for source_path in self._source_paths], [target_path[idx] for
                                                                                                target_path in
                                                                                                self._target_paths]
        sample = Sample(x=source_image, y=target_image, dataset_id=self._dataset_id, is_labeled=True)

        if self._transform is not None:
            sample = self._transform(sample)
        return sample


class MultimodalNiftiDataset(MultimodalDataset):
    """
    Create a dataset class in PyTorch for reading N types of NIfTI files to M types of output NIfTI files
    Note that all images must have the same dimensions.
    """

    def __init__(self, source_dirs: List[str], target_dirs: List[str], dataset_id: int = None,
                 transform: Optional[Callable] = None) -> None:
        super(MultimodalNiftiDataset, self).__init__(source_dirs, target_dirs, dataset_id, transform)



#  -*- coding: utf-8 -*-
#  Copyright 2019 SAMITorch Authors. All Rights Reserved.
#  #
#  Licensed under the MIT License;
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://opensource.org/licenses/MIT
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  ==============================================================================

from typing import Tuple


class SliceBuilder(object):
    """
    Build slices for a data set.
    """

    def __init__(self, image_shape: Tuple[int, int, int, int], patch_size: Tuple[int, int, int, int],
                 step: Tuple[int, int, int, int]):
        """

        Args:
            image_shape (tuple of int): The shape of a dataset image.
            patch_size(tuple of int): The size of the patch to produce.
            step (tuple of int): The size of the stride between patches.
        """

        self._image_slices = self._build_slices(image_shape, patch_size, step)

    @property
    def image_slices(self) -> list:
        """
        Image's slices.

        Returns:
            list: List of image's slices.
        """
        return self._image_slices

    @staticmethod
    def _build_slices(image_shape: Tuple[int, int, int, int], patch_size: Tuple[int, int, int, int],
                      step: Tuple[int, int, int, int]) -> list:
        """
        Iterates over a given n-dim dataset patch-by-patch with a given step and builds an array of slice positions.

        Args:
            image_shape (tuple of int): The shape of a dataset image.
            patch_size(tuple of int): The size of the patch to produce.
            step (tuple of int): The size of the stride between patches.

        Returns:
            list: list of slices.
        """
        slices = []
        channels, i_z, i_y, i_x = image_shape
        k_c, k_z, k_y, k_x = patch_size
        s_c, s_z, s_y, s_x = step
        z_steps = SliceBuilder._gen_indices(i_z, k_z, s_z)
        for z in z_steps:
            y_steps = SliceBuilder._gen_indices(i_y, k_y, s_y)
            for y in y_steps:
                x_steps = SliceBuilder._gen_indices(i_x, k_x, s_x)
                for x in x_steps:
                    slice_idx = (
                        slice(z, z + k_z),
                        slice(y, y + k_y),
                        slice(x, x + k_x)
                    )
                    if len(image_shape) == 4:
                        slice_idx = (slice(0, channels),) + slice_idx
                    slices.append(slice_idx)
        return slices

    @staticmethod
    def _gen_indices(i: int, k: int, s: int):
        """
        Generate slice indices.

        Args:
            i (int): image's coordinate.
            k (int): patch size.
            s (int): step size.

        Returns:
            generator: A generator of indices.
        """
        assert i >= k, 'Sample size has to be bigger than the patch size.'
        for j in range(0, i - k + 1, s):
            yield j
        if j + k < i:
            yield i - k



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest

import numpy as np
from hamcrest import *

from samitorch.metrics.gauges import RunningAverageGauge


class AccuracyGaugeTest(unittest.TestCase):
    VALID_VALUE_1 = 0.50
    VALID_VALUE_2 = 0.70
    VALID_VALUE_3 = 0.90
    INVALID_VALUE = "DARTH VADER"

    def setUp(self):
        self.gauge = RunningAverageGauge()
        self.gauge.reset()

    def tearDown(self):
        self.gauge.reset()

    def test_should_keep_a_valid_average_and_reset(self):
        expected_average = np.array([self.VALID_VALUE_1, self.VALID_VALUE_2, self.VALID_VALUE_3]).mean()
        expected_reset_average = 0

        self.gauge.update(self.VALID_VALUE_1)
        self.gauge.update(self.VALID_VALUE_2, 2)
        self.gauge.update(self.VALID_VALUE_3)

        assert_that(self.gauge.average, close_to(expected_average, 0.00001))

        self.gauge.reset()

        assert_that(self.gauge.average, equal_to(expected_reset_average))

    def test_average_should_be_zero_for_empty_gauge(self):
        expected_average = 0

        assert_that(self.gauge.average, equal_to(expected_average))

    def test_should_reset(self):
        assert_that(calling(self.gauge.update).with_args(self.INVALID_VALUE, 1), raises(TypeError))


if __name__ == 'main':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch

from samitorch.factories.enums import *
from samitorch.factories.factories import ActivationLayerFactory, NormalizationLayerFactory
from samitorch.configs.configurations import ModelConfiguration


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """
    3x3 convolution with zero padding

    Args:
        in_planes (int): The number of input channels.
        out_planes (int): The number of output channels.
        stride (int): The convolution stride. Controls the stride for the cross-correlation
        groups (int): The number of groups in the convolution. Controls the connections between input and outputs.
        dilation (int): Controls the spacing between the kernel points

    Returns:
        :obj:`torch.nn.Conv3d`: A convolution layer.
    """
    return torch.nn.Conv3d(in_planes, out_planes, kernel_size=3, stride=stride,
                           padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """
    1x1 convolution without padding.

    Args:
        in_planes (int): The number of input channels.
        out_planes (int): The number of output channels.
        stride (int): The convolution stride. Controls the stride for the cross-correlation

    Returns:
        :obj:`torch.nn.Conv3d`: A convolution layer.
    """

    return torch.nn.Conv3d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(torch.nn.Module):
    """
    A basic ResNet block.
    """
    expansion = 1

    def __init__(self, in_planes, out_planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, activation="ReLU", norm_num_groups=8):
        """
        Basic block initializer.
        Args:
            in_planes (int): The number of input channels.
            out_planes (int): The number of output channels.
            stride (int): The convolution stride. Controls the stride for the cross-correlation
            downsample (:obj:`torch.nn.Sequential`): A sequential downsampling sub-layer.
            groups (int): The number of groups in the convolution. Controls the connections between input and outputs.
            base_width (int): The base number of output channels.
            dilation (int): Controls the spacing between the kernel points
            activation (str): Desired non-linear activation function.
            norm_num_groups (int): The number of groups for group normalization.
        """
        super(BasicBlock, self).__init__()
        self._activation_layer_factory = ActivationLayerFactory()
        self._normalization_layer_factory = NormalizationLayerFactory()

        if norm_num_groups is not None:
            self._norm1 = self._normalization_layer_factory.create_layer(NormalizationLayers.GroupNorm,
                                                                         norm_num_groups, out_planes)
            self._norm2 = self._normalization_layer_factory.create_layer(NormalizationLayers.GroupNorm,
                                                                         norm_num_groups, out_planes)
        else:
            self._norm1 = self._normalization_layer_factory.create_layer(NormalizationLayers.BatchNormd3d,
                                                                         out_planes)
            self._norm2 = self._normalization_layer_factory.create_layer(NormalizationLayers.BatchNormd3d,
                                                                         out_planes)

        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self._conv1 = conv3x3(in_planes, out_planes, stride)

        if activation == ActivationLayers.PReLU.name:
            self._activation = self._activation_layer_factory.create_layer(ActivationLayers.PReLU)
        else:
            self._activation = self._activation_layer_factory.create_layer(ActivationLayers.ReLU, inplace=True)

        self._conv2 = conv3x3(out_planes, out_planes)

        self._downsample = downsample
        self._stride = stride

    def forward(self, x):
        identity = x

        out = self._conv1(x)
        out = self._norm1(out)
        out = self._activation(out)

        out = self._conv2(out)
        out = self._norm2(out)

        if self._downsample is not None:
            identity = self._downsample(x)

        out += identity
        out = self._activation(out)

        return out


class Bottleneck(torch.nn.Module):
    """
    Bottleneck ResNet block.
    """
    expansion = 4

    def __init__(self, in_planes, out_planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, activation="ReLU", norm_num_groups=8):
        """
        Bottleneck Block initializer.

        Args:
            in_planes (int): The number of input channels.
            out_planes (int): The number of output channels.
            stride (int): The convolution stride. Controls the stride for the cross-correlation
            downsample (:obj:`torch.nn.Sequential`): A sequential downsampling sub-layer.
            groups (int): The number of groups in the convolution. Controls the connections between input and outputs.
            base_width (int): The base number of output channels.
            dilation (int): Controls the spacing between the kernel points
            activation (str): Desired non-linear activation function.
            norm_num_groups (int): The number of groups for group normalization.
        """
        super(Bottleneck, self).__init__()

        self._activation_layer_factory = ActivationLayerFactory()
        self._normalization_layer_factory = NormalizationLayerFactory()

        width = int(out_planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1

        if norm_num_groups is not None:
            self._norm1 = self._normalization_layer_factory.create_layer(NormalizationLayers.GroupNorm, norm_num_groups,
                                                                         width)
            self._norm2 = self._normalization_layer_factory.create_layer(NormalizationLayers.GroupNorm, norm_num_groups,
                                                                         width)
            self._norm3 = self._normalization_layer_factory.create_layer(NormalizationLayers.GroupNorm, norm_num_groups,
                                                                         out_planes * self.expansion)
        else:
            self._norm1 = self._normalization_layer_factory.create_layer(NormalizationLayers.BatchNormd3d, width)
            self._norm2 = self._normalization_layer_factory.create_layer(NormalizationLayers.BatchNormd3d, width)
            self._norm3 = self._normalization_layer_factory.create_layer(NormalizationLayers.BatchNormd3d,
                                                                         out_planes * self.expansion)

        self._conv1 = conv1x1(in_planes, width)

        self._conv2 = conv3x3(width, width, stride, groups, dilation)

        self._conv3 = conv1x1(width, out_planes * self.expansion)

        if activation == ActivationLayers.PReLU.name:
            self._activation = self._activation_layer_factory.create_layer(ActivationLayers.PReLU)
        else:
            self._activation = self._activation_layer_factory.create_layer(ActivationLayers.ReLU, inplace=True)

        self._downsample = downsample
        self._stride = stride

    def forward(self, x):
        identity = x

        out = self._conv1(x)
        out = self._norm1(out)
        out = self._activation(out)

        out = self._conv2(out)
        out = self._norm2(out)
        out = self._activation(out)

        out = self._conv3(out)
        out = self._norm3(out)

        if self._downsample is not None:
            identity = self._downsample(x)

        out += identity
        out = self._activation(out)

        return out


class ResNet3D(torch.nn.Module):
    """
    3D ResNet adaptation from the original ResNet paper :cite:`2016:He`

    .. bibliography:: resnet.bib

    :encoding: utf
    :cited:
    """

    def __init__(self, block: torch.nn.Module, n_blocks_per_layer: list, config: ModelConfiguration):
        """
        ResNet 3D model initializer.
        Args:
            block (:obj:`torch.nn.Module`): The desired block type (Basic or Bottleneck).
            n_blocks_per_layer (list):
            config:
        """
        super(ResNet3D, self).__init__()
        self._activation_layer_factory = ActivationLayerFactory()

        if config.num_groups is not None:
            norm_layer = torch.nn.GroupNorm
        else:
            norm_layer = torch.nn.BatchNorm3d

        self._norm_layer = norm_layer
        self._num_groups = config.num_groups
        self._activation_fn = config.activation
        self._groups = config.conv_groups
        self._base_width = config.width_per_group
        self._replace_stride_with_dilation = config.replace_stride_with_dilation
        self._inplanes = 64
        self._dilation = 1

        if self._replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        else:
            replace_stride_with_dilation = config.replace_stride_with_dilation

        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(config.replace_stride_with_dilation))

        self._conv1 = torch.nn.Conv3d(config.in_channels, self._inplanes, kernel_size=7, stride=2, padding=3,
                                      bias=False)

        if config.num_groups is not None:
            self._norm1 = norm_layer(self._num_groups, self._inplanes)
        else:
            self._norm1 = norm_layer(self._inplanes)

        if self._activation_fn == ActivationLayers.PReLU.name:
            self._activation = self._activation_layer_factory.create_layer(self._activation_fn)
        else:
            self._activation = self._activation_layer_factory.create_layer(self._activation_fn, inplace=True)

        self._maxpool = torch.nn.MaxPool3d(kernel_size=3, stride=2, padding=1)
        self._layer1 = self._make_layer(block, 64, n_blocks_per_layer[0])
        self._layer2 = self._make_layer(block, 128, n_blocks_per_layer[1], stride=2,
                                        dilate=replace_stride_with_dilation[0])
        self._layer3 = self._make_layer(block, 256, n_blocks_per_layer[2], stride=2,
                                        dilate=replace_stride_with_dilation[1])
        self._layer4 = self._make_layer(block, 512, n_blocks_per_layer[3], stride=2,
                                        dilate=replace_stride_with_dilation[2])
        self._avgpool = torch.nn.AdaptiveAvgPool3d((1, 1, 1))
        self._fc = torch.nn.Linear(512 * block.expansion, config.out_channels)

        for m in self.modules():
            if isinstance(m, torch.nn.Conv3d):
                torch.nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (torch.nn.BatchNorm3d, torch.nn.GroupNorm)):
                torch.nn.init.constant_(m.weight, 1)
                torch.nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if config.zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    torch.nn.init.constant_(m._norm3.weight, 0)
                elif isinstance(m, BasicBlock):
                    torch.nn.init.constant_(m.norm.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        num_groups = self._num_groups
        downsample = None
        previous_dilation = self._dilation
        if dilate:
            self._dilation *= stride
            stride = 1

        if num_groups is not None:
            if stride != 1 or self._inplanes != planes * block.expansion:
                downsample = torch.nn.Sequential(
                    conv1x1(self._inplanes, planes * block.expansion, stride),
                    norm_layer(num_groups, planes * block.expansion),
                )
        else:
            if stride != 1 or self._inplanes != planes * block.expansion:
                downsample = torch.nn.Sequential(
                    conv1x1(self._inplanes, planes * block.expansion, stride),
                    norm_layer(planes * block.expansion),
                )
        layers = []
        layers.append(block(self._inplanes, planes, stride, downsample, self._groups,
                            self._base_width, previous_dilation, self._activation_fn, num_groups))
        self._inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self._inplanes, planes, groups=self._groups,
                                base_width=self._base_width, dilation=self._dilation, activation=self._activation_fn))

        return torch.nn.Sequential(*layers)

    def forward(self, x):
        x = self._conv1(x)
        x = self._norm1(x)
        x = self._activation(x)
        x = self._maxpool(x)

        x = self._layer1(x)
        x = self._layer2(x)
        x = self._layer3(x)
        x = self._layer4(x)

        x = self._avgpool(x)
        x = x.reshape(x.size(0), -1)
        x = self._fc(x)

        return x


def _ResNet(block, layers, config):
    """
    Construct the ResNet 3D network.

    Args:
        block (:obj:`torch.nn.Module`): The kind of ResNet block to build the ResNet with.
        layers (list): A list of number of blocks per layer.
        config (:obj:`samitorch.config.model.ModelConfiguration`): The network configuration.

    Returns:
        torch.nn.Module: The ResNet 3D Network.
    """
    model = ResNet3D(block, layers, config)

    return model


def ResNet18(config):
    """Constructs a ResNet-18 model.

    Args:
        config (:obj:`samitorch.config.model.ModelConfiguration`): The network configuration.
    """
    return _ResNet(BasicBlock, [2, 2, 2, 2], config)


def ResNet34(config):
    """Constructs a ResNet-34 model.

    Args:
        config (:obj:`samitorch.config.model.ModelConfiguration`): The network configuration.
    """
    return _ResNet(BasicBlock, [3, 4, 6, 3], config)


def ResNet50(config):
    """Constructs a ResNet-50 model.

    Args:
        config (:obj:`samitorch.config.model.ModelConfiguration`): The network configuration.
    """
    return _ResNet(Bottleneck, [3, 4, 6, 3], config)


def ResNet101(config):
    """Constructs a ResNet-101 model.

    Args:
        config (:obj:`samitorch.config.model.ModelConfiguration`): The network configuration.
    """
    return _ResNet(Bottleneck, [3, 4, 23, 3], config)


def ResNet152(config):
    """Constructs a ResNet-152 model.

    Args:
        config (:obj:`samitorch.config.model.ModelConfiguration`): The network configuration.
    """
    return _ResNet(Bottleneck, [3, 8, 36, 3], config)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import torch
import unittest

from samitorch.factories.parsers import ModelConfigurationParserFactory
from samitorch.models.resnet3d import ResNet3D
from samitorch.factories.factories import ModelFactory
from samitorch.factories.enums import ResNetModels

from tests.models.model_helper_test import TestModelHelper


class ResNet3DTest(unittest.TestCase):
    CONFIGURATION_PATH = "samitorch/configs/resnet3d.yaml"

    def setUp(self):
        self.configurationParserFactory = ModelConfigurationParserFactory()
        self.model_factory = ModelFactory()
        self.config = self.configurationParserFactory.parse(self.CONFIGURATION_PATH)
        self.input = torch.rand((2, 1, 32, 32, 32))
        self.y = torch.randint(low=0, high=2, size=(2, 1)).float()

    def test_model_should_be_created_with_config(self):
        model = self.model_factory.create_model(ResNetModels.ResNet101, self.config)
        assert isinstance(model, ResNet3D)

    def test_model_should_update_vars(self):
        model = self.model_factory.create_model(ResNetModels.ResNet101, self.config)
        helper = TestModelHelper(model, torch.nn.BCEWithLogitsLoss(),
                                 torch.optim.SGD(model.parameters(), lr=0.01))
        helper.assert_vars_change((self.input, self.y))



#  -*- coding: utf-8 -*-
#  Copyright 2019 SAMITorch Authors. All Rights Reserved.
#  #
#  Licensed under the MIT License;
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://opensource.org/licenses/MIT
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  ==============================================================================

import numpy as np
import torch
from visdom import Visdom


class LinePlot(object):
    def __init__(self, visdom, x_label, y_label, title, legend):
        self._visdom = visdom
        self._window = self._visdom.line(X=torch.zeros((1,)).cpu(),
                                         Y=torch.zeros(1).cpu(),
                                         opts=dict(xlabel=x_label,
                                                   ylabel=y_label,
                                                   title=title,
                                                   legend=legend))

    def append(self, x, y):
        self._visdom.line(
            X=x,
            Y=y,
            win=self._window,
            update='append')


class ImagesPlot(object):
    def __init__(self, visdom: Visdom, title):
        self._visdom = visdom
        self._title = title
        self._window = self._visdom.images(tensor=torch.Tensor().new_ones((600, 600)), nrow=4, opts=dict(title=title))

    def update(self, tensor):
        tensor = self._normalize(tensor)

        self._visdom.images(tensor=tensor, nrow=4, win=self._window, opts=dict(title=self._title))

    @staticmethod
    def _normalize(img):
        return (img - np.min(img)) / (np.ptp(img) + 1e-6)


class AccuracyPlot(LinePlot):
    def __init__(self, visdom, title):
        super().__init__(visdom, 'step', 'accuracy', title, ['accuracy'])


class LossPlot(LinePlot):
    def __init__(self, visdom, title):
        super().__init__(visdom, 'step', 'loss', title, ['loss'])


class ParameterPlot(LinePlot):
    def __init__(self, visdom, title, parameter_name):
        super().__init__(visdom, 'epoch', parameter_name, title, [parameter_name])



#  -*- coding: utf-8 -*-
#  Copyright 2019 SAMITorch Authors. All Rights Reserved.
#  #
#  Licensed under the MIT License;
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://opensource.org/licenses/MIT
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  ==============================================================================

from samitorch.inputs.batch import Batch


def sample_collate(batch: list):
    batch = Batch(samples=batch)
    return batch



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import unittest
import torch
import numpy as np

from hamcrest import *
from losses.losses import DiceLoss, GeneralizedDiceLoss, WeightedCrossEntropyLoss
from samitorch.utils.utils import to_onehot


def get_y_true_y_pred():
    # Generate an image with labels 0 (background), 1, 2
    # 3 classes:
    y_true = np.zeros((30, 30), dtype=np.int)
    y_true[1:11, 1:11] = 1
    y_true[15:25, 15:25] = 2

    y_pred = np.zeros((30, 30), dtype=np.int)
    y_pred[5:15, 1:11] = 1
    y_pred[20:30, 20:30] = 2
    return y_true, y_pred


def compute_tensor_y_true_y_logits(y_true, y_pred):
    # Create torch.tensor from numpy
    y_true_tensor = torch.from_numpy(y_true).unsqueeze(0)
    # Create logits torch.tensor:
    num_classes = max(np.max(y_true), np.max(y_pred)) + 1
    y_probas = np.ones((num_classes,) + y_true.shape) * 0.0
    for i in range(num_classes):
        y_probas[i, (y_pred == i)] = 1.0
    y_logits = torch.from_numpy(y_probas).unsqueeze(0).type(torch.float32)
    return y_true_tensor, y_logits


def compute_dice_truth(y_true, y_pred):
    true_res = [0, 0, 0]
    for index in range(3):
        bin_y_true = y_true == index
        bin_y_pred = y_pred == index
        intersection = bin_y_true & bin_y_pred
        true_res[index] = 2.0 * intersection.sum() / (bin_y_pred.sum() + bin_y_true.sum())
    return true_res


def compute_generalized_dice_loss_truth(y_true, y_pred):
    true_res = [0, 0, 0]
    for index in range(3):
        bin_y_true = y_true == index
        bin_y_pred = y_pred == index
        weights = (1.0 / (np.sum(bin_y_true) * np.sum(bin_y_true) + 1e-15))
        intersection = (bin_y_true & bin_y_pred)
        true_res[index] = 2 * intersection.sum() * weights / (((bin_y_pred.sum() + bin_y_true.sum()) * weights) + 1e-15)
    return true_res


class TestDiceLoss(unittest.TestCase):
    INVALID_VALUE_1 = -1
    INVALID_VALUE_2 = "STEVE JOBS"
    INVALID_VALUE_3 = 10
    INVALID_VALUE_4 = 11
    INVALID_REDUCTION = "sum"

    def setUp(self):
        self.y_true, self.y_pred = get_y_true_y_pred()
        self.y_true_tensor, self.y_logits = compute_tensor_y_true_y_logits(self.y_true, self.y_pred)
        self.dice = compute_dice_truth(self.y_true, self.y_pred)
        self.mean_dice_loss = np.subtract(1.0, np.mean(self.dice))

    def test_should_raise_exception_with_bad_values(self):
        assert_that(calling(DiceLoss).with_args(reduction=self.INVALID_REDUCTION), raises(AssertionError))

        dice_loss = DiceLoss()

        assert_that(calling(dice_loss.forward).with_args(inputs=None, targets=None),
                    raises(AttributeError))
        assert_that(calling(dice_loss.forward).with_args(inputs=self.y_logits, targets=None),
                    raises(AttributeError))
        assert_that(calling(dice_loss.forward).with_args(inputs=None, targets=self.y_true_tensor),
                    raises(AttributeError))
        assert_that(calling(dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                         ignore_index=self.INVALID_VALUE_1), raises(AssertionError))
        assert_that(calling(dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                         ignore_index=self.INVALID_VALUE_2), raises(TypeError))
        assert_that(calling(dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                         ignore_index=self.INVALID_VALUE_3), raises(AssertionError))
        assert_that(calling(dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                         ignore_index=self.INVALID_VALUE_4), raises(AssertionError))

    def test_should_compute_dice(self):
        dice_loss = DiceLoss(reduction=None)
        loss = dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3))

        np.testing.assert_almost_equal(loss.numpy(), np.subtract(1.0, self.dice))

    def test_should_compute_dice_for_multiclass_with_ignored_index(self):
        for ignore_index in range(3):
            dice_loss = DiceLoss(reduction=None)
            res = dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3),
                                    ignore_index=ignore_index)
            true_res = np.subtract(1.0, self.dice[:ignore_index] + self.dice[ignore_index + 1:])
            np.testing.assert_almost_equal(res.numpy(), true_res), "{}: {} vs {}".format(ignore_index, res, true_res)

    def test_should_compute_mean_dice(self):
        dice_loss = DiceLoss()
        loss = dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3))

        np.testing.assert_almost_equal(loss.numpy(), self.mean_dice_loss)

    def test_should_compute_mean_dice_for_multiclass_with_ignored_index(self):
        for ignore_index in range(3):
            dice_loss = DiceLoss()
            res = dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3),
                                    ignore_index=ignore_index)
            true_res = np.subtract(1.0, self.dice[:ignore_index] + self.dice[ignore_index + 1:])
            np.testing.assert_almost_equal(res.numpy(), true_res), "{}: {} vs {}".format(ignore_index, res, true_res)


class TestGeneralizedDiceLoss(unittest.TestCase):
    INVALID_VALUE_1 = -1
    INVALID_VALUE_2 = "STEVE JOBS"
    INVALID_VALUE_3 = 10
    INVALID_VALUE_4 = 11
    INVALID_REDUCTION = "sum"

    def setUp(self):
        self.y_true, self.y_pred = get_y_true_y_pred()
        self.y_true_tensor, self.y_logits = compute_tensor_y_true_y_logits(self.y_true, self.y_pred)
        self.generalized_dice_loss = compute_generalized_dice_loss_truth(self.y_true, self.y_pred)
        self.mean_generalized_dice_loss = np.subtract(1.0, np.mean(self.generalized_dice_loss))

    def test_should_raise_exception_with_bad_values(self):
        assert_that(calling(GeneralizedDiceLoss).with_args(reduction=self.INVALID_REDUCTION), raises(AssertionError))

        generalized_dice_loss = GeneralizedDiceLoss()

        assert_that(calling(generalized_dice_loss.forward).with_args(inputs=None, targets=None),
                    raises(AttributeError))
        assert_that(calling(generalized_dice_loss.forward).with_args(inputs=self.y_logits, targets=None),
                    raises(AttributeError))
        assert_that(calling(generalized_dice_loss.forward).with_args(inputs=None, targets=self.y_true_tensor),
                    raises(AttributeError))
        assert_that(calling(generalized_dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                     ignore_index=self.INVALID_VALUE_1),
                    raises(AssertionError))
        assert_that(calling(generalized_dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                     ignore_index=self.INVALID_VALUE_2),
                    raises(TypeError))
        assert_that(calling(generalized_dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                     ignore_index=self.INVALID_VALUE_3),
                    raises(AssertionError))
        assert_that(calling(generalized_dice_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                     ignore_index=self.INVALID_VALUE_4),
                    raises(AssertionError))

    def test_should_compute_generalized_dice(self):
        generalized_dice_loss = GeneralizedDiceLoss()
        loss = generalized_dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3))
        np.testing.assert_almost_equal(loss.numpy(), self.mean_generalized_dice_loss)

    def test_should_compute_generalized_dice_for_multiclass_with_ignored_index(self):
        for ignore_index in range(3):
            generalized_dice_loss = GeneralizedDiceLoss()
            res = generalized_dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3),
                                                ignore_index=ignore_index)
            true_res = np.subtract(1.0, self.generalized_dice_loss[:ignore_index] + self.generalized_dice_loss[ignore_index + 1:])
            np.testing.assert_almost_equal(res.numpy(), true_res), "{}: {} vs {}".format(ignore_index, res, true_res)

    def test_should_compute_mean_generalized_dice(self):
        dice_loss = GeneralizedDiceLoss()
        loss = dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3))

        np.testing.assert_almost_equal(loss.numpy(), self.mean_generalized_dice_loss)

    def test_should_compute_mean_generalized_dice_for_multiclass_with_ignored_index(self):
        for ignore_index in range(3):
            dice_loss = GeneralizedDiceLoss()
            res = dice_loss.forward(self.y_logits, to_onehot(self.y_true_tensor, num_classes=3),
                                    ignore_index=ignore_index)
            true_res = np.subtract(1.0, self.generalized_dice_loss[:ignore_index] + self.generalized_dice_loss[
                                                                                    ignore_index + 1:])
            np.testing.assert_almost_equal(res.numpy(), true_res), "{}: {} vs {}".format(ignore_index, res, true_res)


class TestWeightedCrossEntropy(unittest.TestCase):
    INVALID_VALUE_1 = -1
    INVALID_VALUE_2 = "STEVE JOBS"
    INVALID_VALUE_3 = 10
    INVALID_VALUE_4 = 11
    INVALID_REDUCTION = "sum"

    def setUp(self):
        self.y_true, self.y_pred = get_y_true_y_pred()
        self.y_true_tensor, self.y_logits = compute_tensor_y_true_y_logits(self.y_true, self.y_pred)

    def test_should_raise_exception_with_bad_values(self):
        assert_that(calling(WeightedCrossEntropyLoss).with_args(reduction=self.INVALID_REDUCTION),
                    raises(AssertionError))

        weighted_cross_entropy_loss = WeightedCrossEntropyLoss()

        assert_that(calling(weighted_cross_entropy_loss.forward).with_args(inputs=None, targets=None),
                    raises(AttributeError))
        assert_that(calling(weighted_cross_entropy_loss.forward).with_args(inputs=self.y_logits, targets=None),
                    raises(AttributeError))
        assert_that(calling(weighted_cross_entropy_loss.forward).with_args(inputs=None, targets=self.y_true_tensor),
                    raises(AttributeError))
        assert_that(
            calling(weighted_cross_entropy_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                   ignore_index=self.INVALID_VALUE_1),
            raises(AssertionError))
        assert_that(
            calling(weighted_cross_entropy_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                   ignore_index=self.INVALID_VALUE_2),
            raises(TypeError))
        assert_that(
            calling(weighted_cross_entropy_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                   ignore_index=self.INVALID_VALUE_3),
            raises(AssertionError))
        assert_that(
            calling(weighted_cross_entropy_loss.forward).with_args(inputs=self.y_logits, targets=self.y_true_tensor,
                                                                   ignore_index=self.INVALID_VALUE_4),
            raises(AssertionError))

    def test_should_compute_weights(self):
        weighted_cross_entropy_loss = WeightedCrossEntropyLoss()
        weights = weighted_cross_entropy_loss._compute_class_weights(self.y_logits)
        np.testing.assert_almost_equal(weights.numpy(), np.array([0.2857143, 8.0, 8.0]), decimal=7)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest
import torch

from hamcrest import *

from samitorch.inputs.datasets import MultimodalNiftiDataset, NiftiPatchDataset
from samitorch.inputs.dataloaders import DataLoader
from samitorch.inputs.utils import sample_collate
from torchvision.transforms import Compose
from samitorch.inputs.transformers import ToNDTensor
from samitorch.inputs.batch import Batch


class DataLoaderTest(unittest.TestCase):

    def setUp(self):
        transforms_ = Compose([ToNDTensor()])
        self._dataset = MultimodalNiftiDataset(source_dirs=["tests/data/test_dataset/T1",
                                                            "tests/data/test_dataset/T2"],
                                               target_dirs=["tests/data/test_dataset/label"])

        self._patch_dataset = NiftiPatchDataset(source_dir="tests/data/test_dataset/T1",
                                                target_dir="tests/data/test_dataset/label",
                                                patch_shape=(1, 32, 32, 32),
                                                step=(1, 8, 8, 8),
                                                dataset_id=0,
                                                transform=transforms_)

    def test_should_initialize_dataloader(self):
        dataloader = DataLoader(dataset=self._dataset,
                                batch_size=2,
                                shuffle=True,
                                validation_split=1,
                                num_workers=2,
                                collate_fn=sample_collate)

        assert_that(dataloader, is_not(None))
        assert_that(dataloader, instance_of(torch.utils.data.dataloader.DataLoader))

    def test_should_initialize_dataloader_with_validation_split(self):
        dataloader = DataLoader(dataset=self._dataset,
                                batch_size=2,
                                shuffle=True,
                                validation_split=1,
                                num_workers=2,
                                collate_fn=sample_collate)
        valid_dataloader = dataloader.get_validation_dataloader()

        assert_that(dataloader, is_not(None))
        assert_that(valid_dataloader, is_not(None))
        assert_that(dataloader, instance_of(torch.utils.data.dataloader.DataLoader))
        assert_that(valid_dataloader, instance_of(torch.utils.data.dataloader.DataLoader))
        assert valid_dataloader.sampler.indices not in dataloader.sampler.indices

    def test_should_initialize_patch_dataloader(self):
        dataloader = DataLoader(dataset=self._patch_dataset,
                                batch_size=4,
                                shuffle=True,
                                validation_split=1,
                                num_workers=2,
                                collate_fn=sample_collate)

        assert_that(dataloader, is_not(None))
        assert_that(dataloader, instance_of(torch.utils.data.dataloader.DataLoader))

        batch = next(iter(dataloader))

        assert_that(batch, instance_of(Batch))



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""
Class to describe a Sample object, which contains a source element associated to a ground truth.
"""


class Sample(object):

    def __init__(self, x=None, y=None, template=None, dataset_id: int = None, is_labeled: bool = False):
        """
        Sample initializer.

        Args:
            x (any): Source element.
            y (any): Associated ground truth.
            template (any): A template image (for resampling purpose).
            dataset_id (int): The data set identification number (in case of multi-dataset learning). Default: 0
            is_labeled (bool): Whether or not the sample has a label image (y).
        """
        self._x = x
        self._y = y
        self._template = template
        self._dataset_id = dataset_id
        self._is_labeled = is_labeled

    @property
    def x(self):
        """
        The `x` property of the Sample (generally an image).

        Returns:
            The `x` property.
        """
        return self._x

    @property
    def y(self):
        """
        The `y` property of the Sample (generally an image).

        Returns:
           The `y` property.
        """
        return self._y

    @property
    def template(self):
        """
        The `template` property of the Sample (generally an image from which a transformer is based to resample the `x` image).

        Returns:
           The `template` property.
        """
        return self._template

    @property
    def dataset_id(self):
        """
        A label identifying from which data set the sample is coming from
        Returns:

        """
        return self._dataset_id

    @property
    def is_labeled(self):
        """
        Define if Sample is labeled.

        Returns:
            bool: True if sample is labeled (has a non-None `y` property, else False.
        """
        return self._is_labeled

    @x.setter
    def x(self, x):
        self._x = x

    @y.setter
    def y(self, y):
        self._y = y

    @template.setter
    def template(self, template):
        self._template = template

    @is_labeled.setter
    def is_labeled(self, is_labeled):
        self._is_labeled = is_labeled

    @dataset_id.setter
    def dataset_id(self, dataset_id):
        self._dataset_id = dataset_id

    def update(self, sample):
        """
        Update an existing sample from another Sample.

        Args:
            sample (:obj:`samitorch.inputs.sample.Sample`): Takes the properties of this Sample to update an existing
                Sample.

        Returns:
            :obj:`samitorch.inputs.sample.Sample`: The updated Sample.
        """
        self._x = sample.x
        self._y = sample.y
        self._dataset_id = sample.dataset_id
        self._is_labeled = sample.is_labeled
        return self

    def unpack(self) -> tuple:
        """
        Unpack a Sample.

        Returns:
            tuple: A Tuple of elements representing the (X, y) properties of the Sample.
        """
        return self._x, self._y

    @classmethod
    def from_sample(cls, sample):
        """
        Create a new Sample from an existing Sample passed in parameter.

        Args:
            sample (:obj:`samitorch.inputs.sample.Sample`): A template Sample.

        Returns:
            :obj:`samitorch.inputs.sample.Sample`: A new Sample object with same properties as the one passed in
                parameter.
        """
        return cls(sample.x, sample.y, sample.template, sample.dataset_id, sample.is_labeled)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest

from hamcrest import *

from samitorch.factories.parsers import ModelConfigurationParserFactory

from samitorch.configs.configurations import UNetModelConfiguration


class ConfigurationParserFactoryTest(unittest.TestCase):
    VALID_CONFIGURATION_PATH = "samitorch/configs/unet3d.yaml"
    INVALID_CONFIGURATION_PATH = "samitorch/configs/invalid.yaml"
    INCORRECT_MODEL_TYPE_CONFIGURATION_PATH = "invalid_model_type_configuration.yaml"
    INCORRECT_MODEL_PROPERTY_CONFIGURATION_PATH = "invalid_model_property_configuration.yaml"

    def setUp(self):
        self.configuration_parser_factory = ModelConfigurationParserFactory()

    def test_instantiating_model_configuration_should_success_with_valid_path(self):
        model_configuration = self.configuration_parser_factory.parse(self.VALID_CONFIGURATION_PATH)
        assert_that(model_configuration, is_not(None))
        assert_that(model_configuration, instance_of(UNetModelConfiguration))

    def test_instantiating_model_configuration_should_fail_with_invalid_path(self):
        assert_that(calling(self.configuration_parser_factory.parse).with_args(self.INVALID_CONFIGURATION_PATH),
                    raises(FileNotFoundError))

    def test_instantiating_model_configuration_should_fail_with_incomplete_yaml_file(self):
        assert_that(
            calling(self.configuration_parser_factory).with_args(self.INCORRECT_MODEL_TYPE_CONFIGURATION_PATH)), raises(
            ValueError)

    def test_instantionating_model_configuration_should_fail_with_incorrect_model_property(self):
        assert_that(
            calling(self.configuration_parser_factory.parse).with_args(
                self.INCORRECT_MODEL_PROPERTY_CONFIGURATION_PATH)), raises(
            KeyError)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A base callback class.

Declare methods a Callback must have.
"""
import abc


class Callback(object):
    """Implements abstract callback interface.

    All callbacks should be derived from this class

    See Also:
        class:`Trainer`
    """

    def __init__(self, *args, **kwargs):
        """ Class initializer.
        Args:
            *args: positional arguments
            **kwargs: keyword arguments
        """
        pass

    @abc.abstractmethod
    def at_epoch_begin(self, **kwargs):
        """Function which will be executed at begin of each epoch

        Args:
            **kwargs: additional keyword arguments

        Returns:
            dict: modified trainer attributes, where the name must correspond to the trainer's attribute name
        """
        return {}

    @abc.abstractmethod
    def at_epoch_end(self, **kwargs):
        """Function which will be executed at end of each epoch

        Args:
            **kwargs: additional keyword arguments

        Returns:
            dict: modified trainer attributes, where the name must correspond to the trainer's attribute name
        """
        return {}



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A base trainer class.

Declare methods a Trainer object must have.
"""

import abc
import torch

from typing import List, Optional
from samitorch.callbacks.callbacks import Callback
from samitorch.training.trainer_configuration import TrainerConfiguration
from samitorch.inputs.batch import Batch


class Trainer(object):
    def __init__(self, config: TrainerConfiguration, callbacks: Optional[List[Callback]]):
        """Class initializer.

        Args:
            config (:obj:`samitorch.training.training_config.TrainingConfig`): A TrainingConfig containing training configuration.
            callbacks (:obj:`list` of :obj:`Callback`): A list of Callback objects to register.
        """
        assert config is not None, "Training must at least have a list of one configuration."
        self._config = None

        self._register_config(config)

        if callbacks is not None:
            self._callbacks = list()
            for cbck in callbacks:
                self._register_callback(cbck)

        self._global_step = torch.Tensor().new_zeros((1,), dtype=torch.int64, device='cpu')
        self._epoch = torch.Tensor().new_zeros((1,), dtype=torch.int64, device='cpu')

    @property
    def global_step(self):
        """int: The current epoch count."""
        return self._global_step

    @global_step.setter
    def global_step(self, global_step):
        self._global_step = global_step

    @property
    def epoch(self):
        """int: The current epoch count."""
        return self._epoch

    @epoch.setter
    def epoch(self, epoch):
        self._epoch = epoch

    @property
    def config(self):
        """:obj:`list` of :obj:`TrainingConfig`: A list of registered training configuration, one per model."""
        return self._config

    @property
    def callbacks(self):
        """:obj:`list` of :obj:`Callback`: A list of registered callbacks."""
        return self._callbacks

    @abc.abstractmethod
    def train(self):
        """Main training loop.

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _train_epoch(self, epoch_num: int, *args, **kwargs):
        """Train a model for one epoch.

        Args:
            epoch_num (int): current epoch number.
            **kwargs (dict): additional keyword arguments.

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _train_batch(self, batch: Batch, *args, **kwargs):
        """Function which handles prediction from batch, logging, loss calculation and optimizer step.

        This function is needed for the framework to provide a generic trainer function which works with all kind of
        networks and loss functions. The closure function must implement all steps from forwarding, over loss
        calculation, metric calculation, logging, and the actual backpropagation. It is called with an empty
        optimizer-dict to evaluate and should thus work with optional optimizers.

        Args:
            batch (:obj:`samitorch.inputs.batch.Batch`): Batch object containing the data.
            *args (dict): additional arguments.
            **kwargs (dict): additional keyword arguments.

        Returns:
            dict: Metric values (with same keys as input dict metrics).
            dict: Loss values (with same keys as input dict criterions).
            list: Arbitrary number of predictions.

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """
        raise NotImplementedError()

    @staticmethod
    def _prepare_batch(batch: Batch, input_device: torch.device, output_device: torch.device):
        """Converts a numpy batch of data and labels to suitable datatype and pushes them to correct devices

        Args
            batch (dict): dictionary containing the batch (must have keys 'data' and 'label'
            input_device (:obj:`torch.device`): device for network inputs
            output_device (:obj:`torch.device`): device for network outputs

        Returns:
            dict: dictionary containing all necessary data in right format and type and on the correct device

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _validate_epoch(self, epoch_num: int, *args, **kwargs):
        """Run validation phase.

         Args:
            epoch_num (int): Current epoch number.
            kwargs: keyword arguments

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _validate_batch(self, batch: Batch, *args, **kwargs):
        """ Run validation over a batch.

        Args:
            batch (:obj:`samitorch.inputs.batch.Batch`): An input Batch object,
            **kwargs: additional keywords arguments.

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _finalize(self, *args, **kwargs):
        """Finalize all operations (e.g. save checkpoint, finalize Data Loaders and close logger if required).

        Args:
            *args: positional arguments
            **kwargs (dict): additional keyword arguments.

        Raises:
            NotImplementedError: if not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _setup(self, *args, **kwargs):
        """Defines the actual Trainer Setup.

        Args:
            *args: positional arguments
            **kwargs (dict): additional keyword arguments.

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _at_training_begin(self, *args, **kwargs):
        """Defines the behaviour at beginnig of the training.

        Args:
            *args: positional arguments
            **kwargs (dict): additional keyword arguments.

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _at_training_end(self, *args, **kwargs):
        """Defines the behaviour at the end of the training.

        Args:
            *args: positional arguments
            **kwargs (dict): additional keyword arguments.

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _at_epoch_begin(self, *args, **kwargs):
        """Defines the behaviour at beginnig of each epoch.

        Args:
            *args: positional arguments
            **kwargs (dict): additional keyword arguments.

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _at_epoch_end(self, *args, **kwargs):
        """
        Defines the behaviour at the end of each epoch.

        Args:
            *args: positional arguments
            **kwargs(dict): additional keyword arguments.

        Raises:
            NotImplementedError: If not overwritten by subclass.
        """

        raise NotImplementedError()

    @staticmethod
    def _is_better_val_scores(old_val_score: float, new_val_score: float, mode='highest'):
        """Check whether the new val score is better than the old one with respect to the optimization goal.

        Args:
            old_val_score (float): old validation score
            new_val_score (float): new validation score
            mode (str): String to specify whether a higher or lower validation score is optimal;
                must be in ['highest', 'lowest']

        Returns:
            bool: True if new score is better, False otherwise.
        """

        assert mode in ['highest', 'lowest'], "Invalid Comparison Mode"

        if mode == 'highest':
            return new_val_score > old_val_score
        elif mode == 'lowest':
            return new_val_score < old_val_score

    def _register_callback(self, callback: Callback):
        """Register Callback to Trainer.

        Args:
            callback (:class:`Callback`): the callback to register

        Raises:
            AssertionError: `callback` is not an instance of :class:`Callback` and has not both methods
            ['at_epoch_begin', 'at_epoch_end'].
        """

        assertion_str = "Given callback is not valid; Must be instance of " \
                        "Callback or provide functions " \
                        "'at_epoch_begin' and 'at_epoch_end'"

        assert \
            isinstance(callback, Callback) \
            or (hasattr(callback, "at_epoch_begin")
                and hasattr(callback, "at_epoch_end")), assertion_str

        self._callbacks.append(callback)

    def _register_config(self, config: TrainerConfiguration):
        """Register a TrainingConfig to Trainer.

        Args:
            config (:obj:`TrainingConfig`): the training configuration to register.

        Raises:
             AssertionError: `config` is not an instance of :class:`TrainingConfig`.

        """
        assertion_str = "Given config is not valid; Must be instance of 'TrainingConfig'"

        assert isinstance(config, TrainerConfiguration), assertion_str

        self._config = config



import pathlib

from setuptools import find_packages, setup

# The text of the README file
README_CONTENT = (pathlib.Path(__file__).parent / "README.md").read_text()

setup(
    name='SAMITorch',
    version='0.1.51',
    description='Deep Learning Framework For Medical Image Analysis',
    long_description=README_CONTENT,
    long_description_content_type='text/markdown',
    author='Benoit Anctil-Robitaille and Pierre-Luc Delisle',
    author_email='benoit.anctil-robitaille.1@ens.etsmtl.ca',
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=("tests",)),
    install_requires=['numpy>=1.16.1',
                      'torch>=1.1.0',
                      'torchvision>=0.2.2',
                      'nibabel>=2.4.0',
                      'nipype>=1.1.9',
                      'nilearn>=0.5.0',
                      'scikit-learn>=0.20.3',
                      'pytorch-ignite>=0.2.0',
                      'pynrrd>=0.4.0',
                      'opencv-python>=4.1.0']
)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import unittest
import torch
import os
import numpy as np
import nibabel as nib

from torchvision.transforms import transforms

from hamcrest import *

from samitorch.inputs.datasets import NiftiDataset, MultimodalNiftiDataset, NiftiPatchDataset
from samitorch.inputs.transformers import ToNumpyArray, ToNDTensor, ToNifti1Image, NiftiToDisk
from samitorch.inputs.sample import Sample


class NiftiDatasetTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/test_dataset")
    PATH_TO_SOURCE = os.path.join(TEST_DATA_FOLDER_PATH, "T1")
    PATH_TO_TARGET = os.path.join(TEST_DATA_FOLDER_PATH, "label")
    TEST_SOURCE_IMAGE = nib.load(
        os.path.join(TEST_DATA_FOLDER_PATH, "T1/1-T1.nii")).get_fdata()
    TEST_LABEL_IMAGE = nib.load(
        os.path.join(TEST_DATA_FOLDER_PATH, "label/1-Labels.nii")).get_fdata()

    def setUp(self):
        pass

    def test_should_instantiate_non_cached_dataset(self):
        dataset = NiftiDataset(source_dir=self.PATH_TO_SOURCE, target_dir=self.PATH_TO_TARGET)
        assert_that(dataset, is_not(None))

    def test_should_give_a_tuple_of_training_elements(self):
        dataset = NiftiDataset(source_dir=self.PATH_TO_SOURCE, target_dir=self.PATH_TO_TARGET)

        sample = dataset.__getitem__(0)

        np.testing.assert_array_equal(nib.load(sample.x).get_fdata(), self.TEST_SOURCE_IMAGE)
        np.testing.assert_array_equal(nib.load(sample.y).get_fdata(), self.TEST_LABEL_IMAGE)


class MultimodalNiftiDatasetTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/test_dataset")
    PATH_TO_SOURCE_T1 = os.path.join(TEST_DATA_FOLDER_PATH, "T1")
    PATH_TO_SOURCE_T2 = os.path.join(TEST_DATA_FOLDER_PATH, "T2")
    PATHS_TO_SOURCES = [PATH_TO_SOURCE_T1, PATH_TO_SOURCE_T2]
    PATH_TO_TARGET = os.path.join(TEST_DATA_FOLDER_PATH, "label")
    PATHS_TO_TARGETS = [PATH_TO_TARGET]
    TEST_SOURCE_IMAGE_T1 = nib.load(
        os.path.join(TEST_DATA_FOLDER_PATH, "T1/1-T1.nii")).get_fdata()
    TEST_SOURCE_IMAGE_T2 = nib.load(
        os.path.join(TEST_DATA_FOLDER_PATH, "T2/1-T2.nii")).get_fdata()
    TEST_LABEL_IMAGE = nib.load(
        os.path.join(TEST_DATA_FOLDER_PATH, "label/1-Labels.nii")).get_fdata()

    def setUp(self):
        pass

    def test_should_instantiate_non_cached_dataset(self):
        dataset = MultimodalNiftiDataset(source_dirs=self.PATHS_TO_SOURCES, target_dirs=self.PATHS_TO_TARGETS)
        assert_that(dataset, is_not(None))

    def test_should_give_a_tuple_of_training_elements(self):
        dataset = MultimodalNiftiDataset(source_dirs=self.PATHS_TO_SOURCES, target_dirs=self.PATHS_TO_TARGETS)

        sample = dataset.__getitem__(0)

        source_t1, source_t2 = sample.x[0], sample.x[1]
        label = sample.y[0]

        np.testing.assert_array_equal(nib.load(source_t1).get_fdata(), self.TEST_SOURCE_IMAGE_T1)
        np.testing.assert_array_equal(nib.load(source_t2).get_fdata(), self.TEST_SOURCE_IMAGE_T2)
        np.testing.assert_array_equal(nib.load(label).get_fdata(), self.TEST_LABEL_IMAGE)


class NiftiDatasetWithTransformsTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/test_dataset")
    PATH_TO_SOURCE_T1 = os.path.join(TEST_DATA_FOLDER_PATH, "T1")
    PATHS_TO_SOURCES = [PATH_TO_SOURCE_T1]
    PATH_TO_TARGET = os.path.join(TEST_DATA_FOLDER_PATH, "label")
    PATHS_TO_TARGETS = [PATH_TO_TARGET]

    def setUp(self):
        pass

    def test_should_instantiate_dataset_with_transforms(self):
        transforms_ = transforms.Compose([ToNumpyArray()])
        dataset = MultimodalNiftiDataset(source_dirs=self.PATHS_TO_SOURCES, target_dirs=self.PATHS_TO_TARGETS,
                                         transform=transforms_)

        assert_that(dataset, is_not(None))
        assert_that(dataset._transform, is_not(None))

    def test_should_return_a_sample_of_Numpy_ndarrays_with_respective_transform(self):
        transforms_ = transforms.Compose([ToNumpyArray()])
        dataset = MultimodalNiftiDataset(source_dirs=self.PATHS_TO_SOURCES, target_dirs=self.PATHS_TO_TARGETS,
                                         transform=transforms_)
        sample = dataset.__getitem__(0)

        assert_that(sample.x, instance_of(np.ndarray))
        assert_that(sample.x.ndim, is_(4))
        assert_that(sample.x.shape[0], is_(1))
        assert_that(sample.y, instance_of(np.ndarray))
        assert_that(sample.y.shape[0], is_(1))


class MultimodalNiftiDatasetWithTransformsTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/test_dataset")
    PATH_TO_SOURCE_T1 = os.path.join(TEST_DATA_FOLDER_PATH, "T1")
    PATH_TO_SOURCE_T2 = os.path.join(TEST_DATA_FOLDER_PATH, "T2")
    PATHS_TO_SOURCES = [PATH_TO_SOURCE_T1, PATH_TO_SOURCE_T2]
    PATH_TO_TARGET = os.path.join(TEST_DATA_FOLDER_PATH, "label")
    PATHS_TO_TARGETS = [PATH_TO_TARGET]

    def setUp(self):
        pass

    def test_should_instantiate_dataset_with_transforms(self):
        transforms_ = transforms.Compose([ToNumpyArray()])
        dataset = MultimodalNiftiDataset(source_dirs=self.PATHS_TO_SOURCES, target_dirs=self.PATHS_TO_TARGETS,
                                         transform=transforms_)

        assert_that(dataset, is_not(None))
        assert_that(dataset._transform, is_not(None))

    def test_should_return_a_sample_of_Numpy_ndarrays_with_respective_transform(self):
        transforms_ = transforms.Compose([ToNumpyArray()])

        dataset = MultimodalNiftiDataset(source_dirs=self.PATHS_TO_SOURCES, target_dirs=self.PATHS_TO_TARGETS,
                                         transform=transforms_)
        sample = dataset.__getitem__(0)

        assert_that(sample.x, instance_of(np.ndarray))
        assert_that(sample.x.ndim, is_(4))
        assert_that(sample.x.shape[0], is_(2))
        assert_that(sample.y, instance_of(np.ndarray))
        assert_that(sample.y.shape[0], is_(1))


class NiftiPatchDatasetWithTransformsTest(unittest.TestCase):
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/test_dataset")
    PATH_TO_SOURCE = os.path.join(TEST_DATA_FOLDER_PATH, "T1")
    PATH_TO_TARGET = os.path.join(TEST_DATA_FOLDER_PATH, "label")
    TEST_SOURCE_IMAGE = nib.load(
        os.path.join(TEST_DATA_FOLDER_PATH, "T1/1-T1.nii")).get_fdata()
    TEST_LABEL_IMAGE = nib.load(
        os.path.join(TEST_DATA_FOLDER_PATH, "label/1-Labels.nii")).get_fdata()
    TEST_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data")
    OUTPUT_DATA_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "../data/generated/NiftiPatchDataset")
    PATH_TO_SAVE_X = os.path.join(OUTPUT_DATA_FOLDER_PATH, "nifti_patch_x.nii")
    PATH_TO_SAVE_Y = os.path.join(OUTPUT_DATA_FOLDER_PATH, "nifti_patch_y.nii")

    def setUp(self):
        pass

    def test_should_instantiate_dataset_with_transforms(self):
        transforms_ = transforms.Compose([ToNDTensor()])
        dataset = NiftiPatchDataset(source_dir=self.PATH_TO_SOURCE, target_dir=self.PATH_TO_TARGET,
                                    patch_shape=(1, 32, 32, 32), step=(1, 8, 8, 8),
                                    transform=transforms_)

        assert_that(dataset, is_not(None))
        assert_that(dataset._transform, is_not(None))

    def test_should_return_a_sample_of_Numpy_ndarrays_with_respective_transform(self):
        transforms_ = transforms.Compose([ToNDTensor()])
        dataset = NiftiPatchDataset(source_dir=self.PATH_TO_SOURCE, target_dir=self.PATH_TO_TARGET,
                                    patch_shape=(1, 32, 32, 32), step=(1, 32, 32, 32),
                                    transform=transforms_)
        sample = dataset.__getitem__(0)

        assert_that(sample.x, instance_of(torch.Tensor))
        assert_that(sample.x.ndimension(), is_(4))
        assert_that(sample.x.shape[0], is_(1))
        assert_that(sample.y.ndimension(), is_(4))
        assert_that(sample.y, instance_of(torch.Tensor))
        assert_that(sample.y.shape[0], is_(1))

    def test_should_return_a_sample_of_Numpy_ndarrays_for_inspection(self):
        transforms_ = transforms.Compose([ToNDTensor()])
        dataset = NiftiPatchDataset(source_dir=self.PATH_TO_SOURCE, target_dir=self.PATH_TO_TARGET,
                                    patch_shape=(1, 24, 120, 120), step=(1, 24, 120, 120),
                                    transform=transforms_)
        sample = dataset.__getitem__(0)

        sample = Sample(x=sample.x.numpy(), y=sample.y.numpy(), is_labeled=True)

        transforms_ = transforms.Compose(
            [ToNifti1Image([None, None]), NiftiToDisk([self.PATH_TO_SAVE_X, self.PATH_TO_SAVE_Y])])
        transforms_(sample)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import torch
import numpy as np
import unittest

from samitorch.metrics.metrics import compute_dice_coefficient, compute_mean_dice_coefficient, \
    compute_generalized_dice_coefficient, compute_mean_generalized_dice_coefficient, validate_weights_size, \
    validate_num_classes, validate_ignore_index, Dice, GeneralizedDice
from ignite.metrics.confusion_matrix import ConfusionMatrix
from hamcrest import *


def get_y_true_y_pred():
    # Generate an image with labels 0 (background), 1, 2
    # 3 classes:
    y_true = np.zeros((30, 30), dtype=np.int)
    y_true[1:11, 1:11] = 1
    y_true[15:25, 15:25] = 2

    y_pred = np.zeros((30, 30), dtype=np.int)
    y_pred[5:15, 1:11] = 1
    y_pred[20:30, 20:30] = 2
    return y_true, y_pred


def compute_tensor_y_true_y_logits(y_true, y_pred):
    # Create torch.tensor from numpy
    y_true_tensor = torch.from_numpy(y_true).unsqueeze(0)
    # Create logits torch.tensor:
    num_classes = max(np.max(y_true), np.max(y_pred)) + 1
    y_probs = np.ones((num_classes,) + y_true.shape) * -10
    for i in range(num_classes):
        y_probs[i, (y_pred == i)] = 720
    y_logits = torch.from_numpy(y_probs).unsqueeze(0)
    return y_true_tensor, y_logits


def compute_dice_truth(y_true, y_pred):
    true_res = [0, 0, 0]
    for index in range(3):
        bin_y_true = y_true == index
        bin_y_pred = y_pred == index
        intersection = bin_y_true & bin_y_pred
        true_res[index] = 2 * intersection.sum() / (bin_y_pred.sum() + bin_y_true.sum())
    return true_res


def compute_generalized_dice_truth(y_true, y_pred):
    true_res = [0, 0, 0]
    weights = [0, 0, 0]
    for index in range(3):
        bin_y_true = y_true == index
        bin_y_pred = y_pred == index
        weights[index] = (1.0 / (np.sum(bin_y_true) * np.sum(bin_y_true) + 1e-15))
        intersection = (bin_y_true & bin_y_pred)
        true_res[index] = 2 * intersection.sum() * weights[index] / (
                ((bin_y_pred.sum() + bin_y_true.sum()) * weights[index]) + 1e-15)
    return true_res, weights


class TestValidationMethods(unittest.TestCase):
    INVALID_VALUE_1 = -1
    INVALID_VALUE_2 = 10
    INVALID_VALUE_3 = 11
    NUM_CLASSES = 5

    def setUp(self):
        pass

    def test_should_raise_assertion_error_with_bad_values(self):
        assert_that(calling(validate_ignore_index).with_args(self.INVALID_VALUE_1), raises(AssertionError))
        assert_that(
            calling(validate_num_classes).with_args(ignore_index=self.INVALID_VALUE_2, num_classes=self.NUM_CLASSES),
            raises(AssertionError))
        assert_that(calling(validate_weights_size).with_args(weights_size=self.INVALID_VALUE_2,
                                                             num_classes=self.INVALID_VALUE_3), raises(AssertionError))


class TestDiceMetric(unittest.TestCase):
    INVALID_VALUE_1 = -1
    INVALID_VALUE_2 = "STEVE JOBS"
    INVALID_VALUE_3 = 10
    INVALID_VALUE_4 = 11

    def setUp(self):
        self.y_true, self.y_pred = get_y_true_y_pred()
        self.y_true_tensor, self.y_logits = compute_tensor_y_true_y_logits(self.y_true, self.y_pred)
        self.dice_truth = compute_dice_truth(self.y_true, self.y_pred)
        self.mean_dice_truth = np.mean(self.dice_truth)

    def test_should_raise_exception_with_bad_values(self):
        confusion_matrix = ConfusionMatrix(num_classes=10)

        assert_that(calling(compute_dice_coefficient).with_args(cm=None), raises(AttributeError))
        assert_that(calling(compute_dice_coefficient).with_args(cm=confusion_matrix, ignore_index=self.INVALID_VALUE_1),
                    raises(AssertionError))
        assert_that(calling(compute_dice_coefficient).with_args(cm=confusion_matrix, ignore_index=self.INVALID_VALUE_2),
                    raises(TypeError))
        assert_that(calling(compute_dice_coefficient).with_args(cm=confusion_matrix, ignore_index=self.INVALID_VALUE_3),
                    raises(AssertionError))
        assert_that(calling(compute_dice_coefficient).with_args(cm=confusion_matrix, ignore_index=self.INVALID_VALUE_4),
                    raises(AssertionError))

    def test_should_compute_dice_for_multiclass(self):
        cm = ConfusionMatrix(num_classes=3)
        dice_coefficient = compute_dice_coefficient(cm)
        cm.update((self.y_logits, self.y_true_tensor))
        res = dice_coefficient.compute().numpy()
        assert np.all(res == self.dice_truth)

    def test_should_compute_dice_for_multiclass_with_ignored_index(self):
        for ignore_index in range(3):
            cm = ConfusionMatrix(num_classes=3)
            dice_coefficient = compute_dice_coefficient(cm, ignore_index=ignore_index)
            cm.update((self.y_logits, self.y_true_tensor))
            res = dice_coefficient.compute().numpy()
            true_res = self.dice_truth[:ignore_index] + self.dice_truth[ignore_index + 1:]
            assert np.all(res == true_res), "{}: {} vs {}".format(ignore_index, res, true_res)

    def test_should_compute_mean_dice(self):
        cm = ConfusionMatrix(num_classes=3)
        mean_dice_coefficient = compute_mean_dice_coefficient(cm)
        cm.update((self.y_logits, self.y_true_tensor))
        res = mean_dice_coefficient.compute().numpy()
        assert_that(res, equal_to(self.mean_dice_truth))

    def test_should_compute_mean_dice_with_ignored_index(self):
        for ignore_index in range(3):
            cm = ConfusionMatrix(num_classes=3)
            mean_dice_coefficient = compute_mean_dice_coefficient(cm, ignore_index=ignore_index)
            cm.update((self.y_logits, self.y_true_tensor))
            res = mean_dice_coefficient.compute().numpy()
            true_res = np.mean(self.dice_truth[:ignore_index] + self.dice_truth[ignore_index + 1:])
            assert_that(res, equal_to(true_res)), "{}: {} vs {}".format(ignore_index, res, true_res)


class DiceTest(unittest.TestCase):
    def setUp(self):
        self.y_true, self.y_pred = get_y_true_y_pred()
        self.y_true_tensor, self.y_logits = compute_tensor_y_true_y_logits(self.y_true, self.y_pred)
        self.dice_truth = compute_dice_truth(self.y_true, self.y_pred)
        self.mean_dice_truth = np.mean(self.dice_truth)

    def test_should_compute_dice_metric(self):
        self.dice = Dice(num_classes=3)
        self.dice.update((self.y_logits, self.y_true_tensor))
        res = self.dice.compute()
        np.testing.assert_almost_equal(res, self.dice_truth)

    def test_should_compute_dice_metric_with_sample_average(self):
        self.dice = Dice(num_classes=3, average="samples")
        self.dice.update((self.y_logits, self.y_true_tensor))
        res = self.dice.compute()
        np.testing.assert_almost_equal(res, self.dice_truth)

    def test_should_compute_mean_dice_metric(self):
        self.dice = Dice(num_classes=3, reduction="mean")
        self.dice.update((self.y_logits, self.y_true_tensor))
        res = self.dice.compute()
        assert_that(res, instance_of(torch.Tensor))
        assert_that(res.dtype, is_(torch.float64))
        np.testing.assert_almost_equal(res.mean(), self.mean_dice_truth)


class GeneralizedDiceMetricTest(unittest.TestCase):
    def setUp(self):
        self.y_true, self.y_pred = get_y_true_y_pred()
        self.y_true_tensor, self.y_logits = compute_tensor_y_true_y_logits(self.y_true, self.y_pred)
        self.generalized_dice_truth, weights = compute_generalized_dice_truth(self.y_true, self.y_pred)
        self.weights = torch.from_numpy(np.array(weights))
        self.generalized_mean_dice_truth = np.mean(self.generalized_dice_truth)

    def test_should_compute_generalized_dice_for_multiclass(self):
        self.dice = GeneralizedDice(num_classes=3)
        self.dice.update((self.y_logits, self.y_true_tensor), self.weights)
        res = self.dice.compute()
        np.testing.assert_almost_equal(res, self.generalized_dice_truth)

    def test_should_compute_mean_generalized_dice_for_multiclass(self):
        self.dice = GeneralizedDice(num_classes=3, reduction="mean")
        self.dice.update((self.y_logits, self.y_true_tensor), self.weights)
        res = self.dice.compute()
        assert_that(res, instance_of(torch.Tensor))
        assert_that(res.dtype, is_(torch.float64))
        np.testing.assert_almost_equal(res.mean(), self.generalized_mean_dice_truth)


class TestGeneralizedDiceMetric(unittest.TestCase):
    INVALID_VALUE_1 = -1
    INVALID_VALUE_2 = "STEVE JOBS"
    INVALID_VALUE_3 = 10
    INVALID_VALUE_4 = 11

    def setUp(self):
        self.y_true, self.y_pred = get_y_true_y_pred()
        self.y_true_tensor, self.y_logits = compute_tensor_y_true_y_logits(self.y_true, self.y_pred)
        self.generalized_dice_truth, weights = compute_generalized_dice_truth(self.y_true, self.y_pred)
        self.weights = torch.from_numpy(np.array(weights))
        self.generalized_mean_dice_truth = np.mean(self.generalized_dice_truth)

    def test_should_raise_exception_with_bad_values(self):
        confusion_matrix = ConfusionMatrix(num_classes=10)
        assert_that(calling(compute_generalized_dice_coefficient).with_args(cm=None, weights=self.weights),
                    raises(AttributeError))
        assert_that(calling(compute_generalized_dice_coefficient).with_args(cm=confusion_matrix,
                                                                            weights=self.weights,
                                                                            ignore_index=self.INVALID_VALUE_1),
                    raises(AssertionError))
        assert_that(calling(compute_generalized_dice_coefficient).with_args(cm=confusion_matrix,
                                                                            weights=self.weights,
                                                                            ignore_index=self.INVALID_VALUE_2),
                    raises(TypeError))
        assert_that(calling(compute_generalized_dice_coefficient).with_args(cm=confusion_matrix,
                                                                            weights=self.weights,
                                                                            ignore_index=self.INVALID_VALUE_3),
                    raises(AssertionError))
        assert_that(calling(compute_generalized_dice_coefficient).with_args(cm=confusion_matrix,
                                                                            weights=self.weights,
                                                                            ignore_index=self.INVALID_VALUE_4),
                    raises(AssertionError))

    def test_should_compute_dice_for_multiclass(self):
        cm = ConfusionMatrix(num_classes=3)
        generalized_dice_coefficient = compute_generalized_dice_coefficient(cm, weights=self.weights)
        cm.update((self.y_logits, self.y_true_tensor))
        res = generalized_dice_coefficient.compute().numpy()
        assert np.all(res == self.generalized_dice_truth)

    def test_should_compute_dice_for_multiclass_with_ignored_index(self):
        for ignore_index in range(3):
            cm = ConfusionMatrix(num_classes=3)
            generalized_dice_coefficient = compute_generalized_dice_coefficient(cm, self.weights,
                                                                                ignore_index=ignore_index)
            cm.update((self.y_logits, self.y_true_tensor))
            res = generalized_dice_coefficient.compute().numpy()
            true_res = self.generalized_dice_truth[:ignore_index] + self.generalized_dice_truth[ignore_index + 1:]
            assert np.all(res == true_res), "{}: {} vs {}".format(ignore_index, res, true_res)

    def test_should_compute_mean_dice(self):
        cm = ConfusionMatrix(num_classes=3)
        generalized_dice_coefficient = compute_mean_generalized_dice_coefficient(cm, weights=self.weights)
        cm.update((self.y_logits, self.y_true_tensor))
        res = generalized_dice_coefficient.compute().numpy()
        assert_that(res, equal_to(self.generalized_mean_dice_truth))

    def test_should_compute_mean_dice_with_ignored_index(self):
        for ignore_index in range(3):
            cm = ConfusionMatrix(num_classes=3)
            mean_generalized_dice_coefficient = compute_mean_generalized_dice_coefficient(cm, self.weights,
                                                                                          ignore_index=ignore_index)
            cm.update((self.y_logits, self.y_true_tensor))
            res = mean_generalized_dice_coefficient.compute().numpy()
            true_res = np.mean(
                self.generalized_dice_truth[:ignore_index] + self.generalized_dice_truth[ignore_index + 1:])
            assert_that(res, equal_to(true_res)), "{}: {} vs {}".format(ignore_index, res, true_res)



# -*- coding: utf-8 -*-
# Copyright 2019 SAMITorch Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from enum import Enum


class ImageTypes(Enum):
    NIFTI = 'nifti'
    NRRD = 'nrrd'
    ALL = [NIFTI, NRRD]

    @classmethod
    def from_string(cls, name: str):
        for member in cls:
            if member.name == name:
                return member


class Extensions(Enum):
    NIFTI = ".nii"
    NRRD = ".nrrd"

    @classmethod
    def from_string(cls, name: str):
        for member in cls:
            if member.name == name:
                return member


class Modalities(Enum):
    DTI = 'DTI'
    T1 = 'T1'
    T2 = 'T2'
    MD = 'MD'
    FA = 'FA'
    ALL = [DTI, T1, T2, MD, FA]

    @classmethod
    def from_string(cls, name: str):
        for member in cls:
            if member.name == name:
                return member


class Image(object):

    @staticmethod
    def is_processed_dti(file):
        return Image.is_dti(file) and Image.is_processed(file)

    @staticmethod
    def is_unprocessed_dti(file):
        return Image.is_dti(file) and not Image.is_processed(file)

    @staticmethod
    def is_unprocessed_fa(file):
        return Image.is_fa(file) and not Image.is_processed(file)

    @staticmethod
    def is_unprocessed_md(file):
        return Image.is_md(file) and not Image.is_processed(file)

    @staticmethod
    def is_processed_t1(file):
        return Image.is_t1(file) and Image.is_processed(file)

    @staticmethod
    def is_unprocessed_t1(file):
        return Image.is_t1(file) and not Image.is_processed(file)

    @staticmethod
    def is_nifti_t1_mask(file):
        return Image.is_nifti(file) and Image.is_t1(file) and 'mask' in file

    @staticmethod
    def is_nifti(file):
        return Extensions.NIFTI.value in file

    @staticmethod
    def is_nrrd(file):
        return Extensions.NRRD.value in file

    @staticmethod
    def is_t1(file):
        return Modalities.T1 in file

    @staticmethod
    def is_dti(file):
        return Modalities.DTI in file

    @staticmethod
    def is_fa(file):
        return Modalities.FA in file

    @staticmethod
    def is_md(file):
        return Modalities.MD in file

    @staticmethod
    def is_processed(file):
        return "Processed" in file
